/*
** file.c
**
** Copyright (c) 1994-1995 Peter Eriksson <pen@signum.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "phttpd.h"


static hashtable_t *content_types_table = NULL;
static hashtable_t *content_encodings_table = NULL;
static char *default_type = NULL;


static struct options file_cfg_table[] =
{
    { "content-types",     T_HTABLE, &content_types_table,	NULL },
    { "content-encodings", T_HTABLE, &content_encodings_table,	NULL },
    
    { "default-type",	   T_STRING, &default_type,             NULL },
    
    { NULL,                -1,       NULL,			NULL }
};



static int get_content_type(const char *path,
			    const struct stat *sp,
			    char *type,
			    char *encoding)
{
    char *suffix;
    hashentry_t *hep;
    int len;
    

    type[0] = '\0';
    if (encoding)
	encoding[0] = '\0';
    
    if (debug > 2)
	fprintf(stderr, "*** get_content_type(\"%s\", ...)\n", path);
    
    if (content_types_table == NULL)
    {
	if (debug > 2)
	    fprintf(stderr, "\t-> Content type unknown (NULL table).\n");

	strcpy(type, default_type);
	return 0;
    }


    /* Get suffix */
    suffix = strrchr(path, '.');
    if (suffix == NULL)
    {
	strcpy(type, default_type);
	return 0;
    }

    len = 0;
    
    if (encoding && content_encodings_table)
    {
	hep = ht_lookup(content_encodings_table, suffix, len);

	if (hep)
	{
	    strcpy(encoding, hep->data);

	    if (debug > 2)
		fprintf(stderr, "\t -> Encoding found: %s\n", encoding);

	    ht_release(hep);

	    if (suffix > path)
	    {
		char *cp;
		
		suffix--;
		for (cp = suffix-1; cp > path && *cp != '.'; cp--)
		    ;
		
		len = suffix - cp+1 ;
		suffix = cp;
		
		if (debug > 1)
		    fprintf(stderr, "file: new suffix: len = %d, str=%.*s\n",
			    len, len, suffix);
	    }
	}
    }

    hep = ht_lookup(content_types_table, suffix, len);
    if (hep)
    {
	strcpy(type, hep->data);
	ht_release(hep);
	
	if (debug > 2)
	    fprintf(stderr, "\t -> Content-Type found: %s\n", type);

	return 0;
    }
    
    if (debug > 2)
	fprintf(stderr, "\t-> Content type unknown.\n");
    
    strcpy(type, default_type);
    return 0;
}

int pm_init(const char **argv)
{
    char *cfg_path, *cp;
    const char *name = argv[0];
    
    
    if (debug > 1)
	fprintf(stderr, "*** file/pm_init(\"%s\") called ***\n", name);

    cfg_path = s_malloc(strlen(name)+6);
    strcpy(cfg_path, name);
    
    cp = strrchr(cfg_path, '.');
    if (cp && strcmp(cp, ".so") == 0)
	*cp = '\0';
    
    strcat(cfg_path, ".conf");
    if (config_parse_file(cfg_path, file_cfg_table, 0) < 0)
	return -1;
    
    if (config_parse_argv(argv+1, file_cfg_table) < 0)
	return -1;
    
    if (default_type == NULL)
	default_type = s_strdup("unknown");

    return 0;
}


void pm_exit(void)
{
    if (debug > 1)
	fprintf(stderr, "*** file/pm_exit() called ***\n");

    if (content_types_table)
    {
	ht_destroy(content_types_table);
	s_free(content_types_table);
    }

    if (content_encodings_table)
    {
	ht_destroy(content_encodings_table);
	s_free(content_encodings_table);
    }

    if (default_type)
	s_free(default_type);
    
    if (debug > 1)
	fprintf(stderr, "*** file/pm_exit() done ***\n");
}



static int http_get_head(struct connectioninfo *cip)
{
    char typebuf[256], buf[256], encbuf[256], ubuf[1024], *s_since;
    int i, result, head_method = 0;
    int fd = cip->fd;
    struct httpinfo *hip = cip->hip;
    fscentry_t *fep;
    fsinfo_t *fip;
    unsigned int fsc_flags = 0;
    
    
    if (debug > 1)
	fprintf(stderr, "*** file/pm_get() called ***\n");

    if (hip && hip->mip && (hip->mip->pragma_flags & MPF_NOCACHE))
	fsc_flags = FSCF_RELOAD;
    
    fep = fscache_lookup(hip->url, fsc_flags);
    if (fep == NULL)
	return -1;

    if (fep->fip == NULL)
    {
	fscache_release(fep);
	return -1;
    }

    fip = fep->fip;
    
    if (S_ISDIR(fip->sb.st_mode))
    {
	sprintf(ubuf, "%s%s/", server_url, hip->url);

	fscache_release(fep);
	return http_redirect(cip, ubuf, hip->request, NULL, 301);
    }

    if (get_content_type(fip->path, &fip->sb, typebuf, encbuf))
    {
	fscache_release(fep);
	return -1;
    }

    if ((s_since = mime_getheader(hip->mip, "IF-MODIFIED-SINCE", 1)) != NULL)
    {
        struct tm tm_since;
	int t_since;


	if (debug > 2)
	    fprintf(stderr, "Got If-Modified-Since: %s\n", s_since);
	
	t_since = atotm(s_since, &tm_since);
	if (t_since != -1)
	{
	    if (fip->sb.st_mtime <= (time_t) t_since)
	    {
		fscache_release(fep);
	        return http_not_modified(cip);
	    }
	}
	else
	{
	    /* Could not parse the date format - do a string compare */
	    http_time_r(&fip->sb.st_mtime, buf, sizeof(buf));
	    i = strlen(buf);
	    buf[i-1] = '\0';
	    
	    if (strcmp(s_since, buf) == 0)
	    {
		fscache_release(fep);
	        return http_not_modified(cip);
	    }
	}
    }

    if (strcasecmp(hip->method, "HEAD") == 0)
	head_method = 1;
    else
	if (fscache_getdata(fep) != 1)
	{
	    if (debug > 1)
		fprintf(stderr, "fscache_getdata(): failed\n");
	    
	    fscache_release(fep);
	    return -1;
	}
    
    result = 200;
    if (hip->mip != NULL)
    {
	http_sendheaders(fd, cip, result, NULL);
	http_sendlastmodified(fd, fip->sb.st_mtime);
	
	if (typebuf[0])
	    fd_puts2nl("Content-Type: ", typebuf, fd);
	
	if (encbuf[0])
	    fd_puts2nl("Content-Encoding: ", encbuf, fd);

	fd_putsinl("Content-Length: ", (int) fip->sb.st_size, fd);
    }

    
    if (hip->mip != NULL)
	fd_putc('\n', fd);

    if (head_method)
    {
	fscache_release(fep);
	return result;
    }
    
    hip->length = fip->sb.st_size;
    
    if (fd_write(fd, fip->data.file.content, fip->sb.st_size) < 0)
	if (debug)
	{
	    fprintf(stderr, "ERROR on #%d: ", cip->request_no);
	    perror("write");
	}

    fscache_release(fep);
    
    if (debug > 2)
	fprintf(stderr, "*** file/http_get: Returning\n");
    
    return result;
}




static int http_put(struct connectioninfo *cip)
{
    char buf[2048], path[1025];
    struct stat sb;
    int fd_out, len, result;
    int fd = cip->fd;
    struct httpinfo *hip = cip->hip;
    
    
    if (url_expand(hip->url, path, sizeof(path), &sb, NULL, NULL) == NULL)
	return -1;
    
    if (S_ISDIR(sb.st_mode))
    {
	sprintf(path, "%s%s/", server_url, hip->url);
	return http_redirect(cip, path, hip->request, NULL, 301);
    }

    fd_out = fd_open(path, O_WRONLY+O_CREAT, 0644);
    if (fd_out < 0)
	return error_system(cip, "do_putpost(): open()");
    
    while ((len = fd_read(fd, buf, sizeof(buf))) > 0)
    {
	if (fd_write(fd_out, buf, len) != len)
	{
	    fd_close(fd_out);
	    return error_system(cip, "do_putpost(): write()");
	}
    }
    
    hip->length = fd_close(fd_out);
    
    result = 200;
    
    if (hip->mip != NULL)
    {
	http_sendheaders(fd, cip, result, NULL);
	
	http_sendlastmodified(fd, sb.st_mtime);
	fd_puts("Content-Type: text/html\n\n", fd);
    }
    
    html_sysheader(fd, "H2", "Succeded");
    fd_puts("OK\n", fd);
    html_sysfooter(fd);

    return result;
}


int pm_request(struct connectioninfo *cip)
{
    struct httpinfo *hip = cip->hip;
    
    if (strcasecmp(hip->method, "GET") == 0 ||
	strcasecmp(hip->method, "HEAD") == 0)
	return http_get_head(cip);
    else if (strcasecmp(hip->method, "PUT") == 0)
	return http_put(cip);
    else
	return -2;
}
